@@ -1,38 +1,6 @@
-/* ***** BEGIN LICENSE BLOCK *****
- * Version: MPL 1.1/GPL 2.0/LGPL 2.1
- *
- * The contents of this file are subject to the Mozilla Public License Version
- * 1.1 (the "License"); you may not use this file except in compliance with
- * the License. You may obtain a copy of the License at
- * http://www.mozilla.org/MPL/
- *
- * Software distributed under the License is distributed on an "AS IS" basis,
- * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
- * for the specific language governing rights and limitations under the
- * License.
- *
- * The Original Code is the Netscape security libraries.
- *
- * The Initial Developer of the Original Code is
- * Netscape Communications Corporation.
- * Portions created by the Initial Developer are Copyright (C) 1994-2000
- * the Initial Developer. All Rights Reserved.
- *
- * Contributor(s):
- *
- * Alternatively, the contents of this file may be used under the terms of
- * either the GNU General Public License Version 2 or later (the "GPL"), or
- * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
- * in which case the provisions of the GPL or the LGPL are applicable instead
- * of those above. If you wish to allow use of your version of this file only
- * under the terms of either the GPL or the LGPL, and not to allow others to
- * use your version of this file under the terms of the MPL, indicate your
- * decision by deleting the provisions above and replace them with the notice
- * and other provisions required by the GPL or the LGPL. If you do not delete
- * the provisions above, a recipient may use your version of this file under
- * the terms of any one of the MPL, the GPL or the LGPL.
- *
- * ***** END LICENSE BLOCK ***** */
+/* This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 /*
  * This file deals with PKCS #11 passwords and authentication.
  */
@@ -78,12 +46,14 @@ static struct PK11GlobalStruct {
  * succeed if the user is already logged in.
  */
 SECStatus
-pk11_CheckPassword(PK11SlotInfo *slot,char *pw)
+pk11_CheckPassword(PK11SlotInfo *slot,char *pw,PRBool contextSpecific)
 {
     int len = 0;
     CK_RV crv;
     SECStatus rv;
     int64 currtime = PR_Now();
+    PRBool mustRetry;
+    int retry = 0;
 
     if (slot->protectedAuthPath) {
 	len = 0;
@@ -95,27 +65,50 @@ pk11_CheckPassword(PK11SlotInfo *slot,char *pw)
 	len = PORT_Strlen(pw);
     }
 
-    PK11_EnterSlotMonitor(slot);
-    crv = PK11_GETTAB(slot)->C_Login(slot->session,CKU_USER,
+    do {
+	PK11_EnterSlotMonitor(slot);
+	crv = PK11_GETTAB(slot)->C_Login(slot->session,
+		contextSpecific ? CKU_CONTEXT_SPECIFIC : CKU_USER,
 						(unsigned char *)pw,len);
-    slot->lastLoginCheck = 0;
-    PK11_ExitSlotMonitor(slot);
-    switch (crv) {
-    /* if we're already logged in, we're good to go */
-    case CKR_OK:
-	slot->authTransact = PK11_Global.transaction;
-    case CKR_USER_ALREADY_LOGGED_IN:
-	slot->authTime = currtime;
-	rv = SECSuccess;
-	break;
-    case CKR_PIN_INCORRECT:
-	PORT_SetError(SEC_ERROR_BAD_PASSWORD);
-	rv = SECWouldBlock; /* everything else is ok, only the pin is bad */
-	break;
-    default:
-	PORT_SetError(PK11_MapError(crv));
-	rv = SECFailure; /* some failure we can't fix by retrying */
-    }
+	slot->lastLoginCheck = 0;
+	mustRetry = PR_FALSE;
+	PK11_ExitSlotMonitor(slot);
+	switch (crv) {
+	/* if we're already logged in, we're good to go */
+	case CKR_OK:
+		/* TODO If it was for CKU_CONTEXT_SPECIFIC should we do this */
+	    slot->authTransact = PK11_Global.transaction;
+	    /* Fall through */
+	case CKR_USER_ALREADY_LOGGED_IN:
+	    slot->authTime = currtime;
+	    rv = SECSuccess;
+	    break;
+	case CKR_PIN_INCORRECT:
+	    PORT_SetError(SEC_ERROR_BAD_PASSWORD);
+	    rv = SECWouldBlock; /* everything else is ok, only the pin is bad */
+	    break;
+	/* someone called reset while we fetched the password, try again once
+	 * if the token is still there. */
+	case CKR_SESSION_HANDLE_INVALID:
+	case CKR_SESSION_CLOSED:
+	    if (retry++ == 0) {
+		rv = PK11_InitToken(slot,PR_FALSE);
+		if (rv == SECSuccess) {
+		    if (slot->session != CK_INVALID_SESSION) {
+			mustRetry = PR_TRUE;
+		    } else {
+			PORT_SetError(PK11_MapError(crv));
+			rv = SECFailure;
+		    }
+		}
+		break;
+	    }
+	    /* Fall through */
+	default:
+	    PORT_SetError(PK11_MapError(crv));
+	    rv = SECFailure; /* some failure we can't fix by retrying */
+	}
+    } while (mustRetry);
     return rv;
 }
 
@@ -124,7 +117,7 @@ pk11_CheckPassword(PK11SlotInfo *slot,char *pw)
  * we are really checking the password.
  */
 SECStatus
-PK11_CheckUserPassword(PK11SlotInfo *slot,char *pw)
+PK11_CheckUserPassword(PK11SlotInfo *slot, const char *pw)
 {
     int len = 0;
     CK_RV crv;
@@ -141,6 +134,21 @@ PK11_CheckUserPassword(PK11SlotInfo *slot,char *pw)
 	len = PORT_Strlen(pw);
     }
 
+    /*
+     * If the token doesn't need a login, don't try to relogin because the
+     * effect is undefined. It's not clear what it means to check a non-empty
+     * password with such a token, so treat that as an error.
+     */
+    if (!slot->needLogin) {
+        if (len == 0) {
+            rv = SECSuccess;
+        } else {
+            PORT_SetError(SEC_ERROR_BAD_PASSWORD);
+            rv = SECFailure;
+        }
+        return rv;
+    }
+
     /* force a logout */
     PK11_EnterSlotMonitor(slot);
     PK11_GETTAB(slot)->C_Logout(slot->session);
@@ -234,7 +242,7 @@ PK11_HandlePasswordCheck(PK11SlotInfo *slot,void *wincx)
 	    NeedAuth = PR_TRUE;
 	}
     }
-    if (NeedAuth) PK11_DoPassword(slot,PR_TRUE,wincx);
+    if (NeedAuth) PK11_DoPassword(slot,PR_TRUE,wincx,PR_FALSE);
 }
 
 void
@@ -293,7 +301,7 @@ pk11_LoginStillRequired(PK11SlotInfo *slot, void *wincx)
 SECStatus
 PK11_Authenticate(PK11SlotInfo *slot, PRBool loadCerts, void *wincx) {
     if (pk11_LoginStillRequired(slot,wincx)) {
-	return PK11_DoPassword(slot,loadCerts,wincx);
+	return PK11_DoPassword(slot,loadCerts,wincx,PR_FALSE);
     }
     return SECSuccess;
 }
@@ -385,7 +393,7 @@ PK11_VerifyPW(PK11SlotInfo *slot,char *pw)
  * initialize a user PIN Value
  */
 SECStatus
-PK11_InitPin(PK11SlotInfo *slot,char *ssopw, char *userpw)
+PK11_InitPin(PK11SlotInfo *slot, const char *ssopw, const char *userpw)
 {
     CK_SESSION_HANDLE rwsession = CK_INVALID_SESSION;
     CK_RV crv;
@@ -437,11 +445,13 @@ done:
     if (rv == SECSuccess) {
         /* update our view of the world */
         PK11_InitToken(slot,PR_TRUE);
-	PK11_EnterSlotMonitor(slot);
-    	PK11_GETTAB(slot)->C_Login(slot->session,CKU_USER,
+	if (slot->needLogin) {
+	    PK11_EnterSlotMonitor(slot);
+	    PK11_GETTAB(slot)->C_Login(slot->session,CKU_USER,
 						(unsigned char *)userpw,len);
-	slot->lastLoginCheck = 0;
-	PK11_ExitSlotMonitor(slot);
+	    slot->lastLoginCheck = 0;
+	    PK11_ExitSlotMonitor(slot);
+	}
     }
     return rv;
 }
@@ -450,18 +460,21 @@ done:
  * Change an existing user password
  */
 SECStatus
-PK11_ChangePW(PK11SlotInfo *slot,char *oldpw, char *newpw)
+PK11_ChangePW(PK11SlotInfo *slot, const char *oldpw, const char *newpw)
 {
     CK_RV crv;
     SECStatus rv = SECFailure;
-    int newLen;
-    int oldLen;
+    int newLen = 0;
+    int oldLen = 0;
     CK_SESSION_HANDLE rwsession;
 
-    if (newpw == NULL) newpw = "";
-    if (oldpw == NULL) oldpw = "";
-    newLen = PORT_Strlen(newpw);
-    oldLen = PORT_Strlen(oldpw);
+    /* use NULL values to trigger the protected authentication path */
+    if (!slot->protectedAuthPath) {
+	if (newpw == NULL) newpw = "";
+	if (oldpw == NULL) oldpw = "";
+    }
+    if (newpw) newLen = PORT_Strlen(newpw);
+    if (oldpw) oldLen = PORT_Strlen(oldpw);
 
     /* get a rwsession */
     rwsession = PK11_GetRWSession(slot);
@@ -519,7 +532,8 @@ PK11_SetIsLoggedInFunc(PK11IsLoggedInFunc func)
  * of the PKCS 11 module.
  */
 SECStatus
-PK11_DoPassword(PK11SlotInfo *slot, PRBool loadCerts, void *wincx)
+PK11_DoPassword(PK11SlotInfo *slot, PRBool loadCerts, void *wincx,
+			PRBool contextSpecific)
 {
     SECStatus rv = SECFailure;
     char * password;
@@ -588,13 +602,12 @@ PK11_DoPassword(PK11SlotInfo *slot, PRBool loadCerts, void *wincx)
 		break;
 	    }
 	}
-	rv = pk11_CheckPassword(slot,password);
+	rv = pk11_CheckPassword(slot,password,contextSpecific);
 	PORT_Memset(password, 0, PORT_Strlen(password));
 	PORT_Free(password);
 	if (rv != SECWouldBlock) break;
     }
     if (rv == SECSuccess) {
-	rv = pk11_CheckVerifyTest(slot);
 	if (!PK11_IsFriendly(slot)) {
 	    nssTrustDomain_UpdateCachedTokenCerts(slot->nssToken->trustDomain,
 	                                      slot->nssToken);
@@ -606,7 +619,7 @@ PK11_DoPassword(PK11SlotInfo *slot, PRBool loadCerts, void *wincx)
 void PK11_LogoutAll(void)
 {
     SECMODListLock *lock = SECMOD_GetDefaultModuleListLock();
-    SECMODModuleList *modList = SECMOD_GetDefaultModuleList();
+    SECMODModuleList *modList;
     SECMODModuleList *mlp = NULL;
     int i;
 
@@ -616,6 +629,7 @@ void PK11_LogoutAll(void)
     }
 
     SECMOD_GetReadLock(lock);
+    modList = SECMOD_GetDefaultModuleList();
     /* find the number of entries */
     for (mlp = modList; mlp != NULL; mlp = mlp->next) {
 	for (i=0; i < mlp->module->slotCount; i++) {
