@@ -1,38 +1,6 @@
-/* ***** BEGIN LICENSE BLOCK *****
- * Version: MPL 1.1/GPL 2.0/LGPL 2.1
- *
- * The contents of this file are subject to the Mozilla Public License Version
- * 1.1 (the "License"); you may not use this file except in compliance with
- * the License. You may obtain a copy of the License at
- * http://www.mozilla.org/MPL/
- *
- * Software distributed under the License is distributed on an "AS IS" basis,
- * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
- * for the specific language governing rights and limitations under the
- * License.
- *
- * The Original Code is the Netscape security libraries.
- *
- * The Initial Developer of the Original Code is
- * Netscape Communications Corporation.
- * Portions created by the Initial Developer are Copyright (C) 1994-2000
- * the Initial Developer. All Rights Reserved.
- *
- * Contributor(s):
- *
- * Alternatively, the contents of this file may be used under the terms of
- * either the GNU General Public License Version 2 or later (the "GPL"), or
- * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
- * in which case the provisions of the GPL or the LGPL are applicable instead
- * of those above. If you wish to allow use of your version of this file only
- * under the terms of either the GPL or the LGPL, and not to allow others to
- * use your version of this file under the terms of the MPL, indicate your
- * decision by deleting the provisions above and replace them with the notice
- * and other provisions required by the GPL or the LGPL. If you do not delete
- * the provisions above, a recipient may use your version of this file under
- * the terms of any one of the MPL, the GPL or the LGPL.
- *
- * ***** END LICENSE BLOCK ***** */
+/* This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 /*
  * This file PK11Contexts which are  used in multipart hashing, 
  * encryption/decryption, and signing/verication operations.
@@ -249,7 +217,8 @@ static PK11Context *pk11_CreateNewContextInSlot(CK_MECHANISM_TYPE type,
     SECStatus rv;
 	
     PORT_Assert(slot != NULL);
-    if (!slot || (!symKey && operation != CKA_DIGEST)) {
+    if (!slot || (!symKey && ((operation != CKA_DIGEST) || 
+	                      (type == CKM_SKIPJACK_CBC64)))) {
 	PORT_SetError(SEC_ERROR_INVALID_ARGS);
 	return NULL;
     }
@@ -291,6 +260,7 @@ static PK11Context *pk11_CreateNewContextInSlot(CK_MECHANISM_TYPE type,
 	    context->param = (SECItem *)&pk11_null_params;
 	}
     } else {
+	PORT_SetError(SEC_ERROR_INVALID_ARGS);
 	context->param = NULL;
     }
     context->init = PR_FALSE;
@@ -325,15 +295,15 @@ __PK11_CreateContextByRawKey(PK11SlotInfo *slot, CK_MECHANISM_TYPE type,
      PK11Origin origin, CK_ATTRIBUTE_TYPE operation, SECItem *key, 
 						SECItem *param, void *wincx)
 {
-    PK11SymKey *symKey;
-    PK11Context *context;
+    PK11SymKey *symKey = NULL;
+    PK11Context *context = NULL;
 
     /* first get a slot */
     if (slot == NULL) {
 	slot = PK11_GetBestSlot(type,wincx);
 	if (slot == NULL) {
 	    PORT_SetError( SEC_ERROR_NO_MODULE );
-	    return NULL;
+	    goto loser;
 	}
     } else {
 	PK11_ReferenceSlot(slot);
@@ -341,12 +311,17 @@ __PK11_CreateContextByRawKey(PK11SlotInfo *slot, CK_MECHANISM_TYPE type,
 
     /* now import the key */
     symKey = PK11_ImportSymKey(slot, type, origin, operation,  key, wincx);
-    if (symKey == NULL) return NULL;
+    if (symKey == NULL) goto loser;
 
     context = PK11_CreateContextBySymKey(type, operation, symKey, param);
 
-    PK11_FreeSymKey(symKey);
-    PK11_FreeSlot(slot);
+loser:
+    if (symKey) {
+        PK11_FreeSymKey(symKey);
+    }
+    if (slot) {
+        PK11_FreeSlot(slot);
+    }
 
     return context;
 }
@@ -618,7 +593,7 @@ SECStatus PK11_DigestBegin(PK11Context *cx)
 
 SECStatus
 PK11_HashBuf(SECOidTag hashAlg, unsigned char *out, unsigned char *in, 
-								int32 len) {
+								PRInt32 len) {
     PK11Context *context;
     unsigned int max_length;
     unsigned int out_length;
@@ -686,7 +661,7 @@ PK11_CipherOp(PK11Context *context, unsigned char * out, int *outlen,
 
     /*
      * The fortezza hack is to send 8 extra bytes on the first encrypted and
-     * loose them on the first decrypt.
+     * lose them on the first decrypt.
      */
     if (context->fortezzaHack) {
 	unsigned char random[8];
